import pkg_resources

from mountepy.http_service import HttpService, ServiceGroup, wait_for_port
from mountepy.mountebank import Mountebank, HttpService, HttpStub

__version__ = pkg_resources.get_distribution(__name__).version



"""
Management of Mountebank installation.
"""

import logging
import os
import subprocess
import tarfile
import urllib


_mb_command = None


def get_mb_command():
    global _mb_command
    if not _mb_command:
        _mb_command = _MountebankInstallManager().get_mb_command()

    return _mb_command


class MBStandaloneBrokenError(Exception):
    """
    Means that the standalone Mountebank setup is broken.
    Probably someone deleted some files from it.
    """
    pass


class _MountebankInstallManager:

    """
    Manages installation of Mountebank.
    Can detect Mountebank installed in the system.
    Can also download and setup a standalone distribution.
    """

    CACHE_DIR = os.path.expanduser('~/.cache/mountepy')
    MB_INSTALL_CHECK_CMD = ['mb', 'help']

    def __init__(self):
        self._log = logging.getLogger(__name__)

    def get_mb_command(self):
        """
        Gets the command to start a Mountebank server in the system.
        Ensures that either Mountebank installation or standalone distribution is available.
        Will trigger Mountebank download if necessary.
        :returns: The command that can be used to start Mountebank server.
        :rtype: list[str]
        """
        if self._check_mb_install(): # pragma: no cover
            mb_command = ['mb']
        else:
            if not os.path.exists(self.CACHE_DIR):
                self._setup_standalone()
            mb_dir = self._get_mb_dir(self.CACHE_DIR)
            node_path = self._get_node_path(mb_dir)
            mb_exe_path = os.path.join(mb_dir, 'mountebank/bin/mb')
            mb_command = [node_path, mb_exe_path]
            self._log.debug('Standalone Mountebank command set to %s', mb_command)
        return mb_command

    def _setup_standalone(self):
        """
        Downloads and sets up a standalone distribution Mountebank.
        Standalone distribution also contains NodeJS.
        """
        self._log.info('Setting up a standalone Mountebank.')
        mb_archive_path, _ = urllib.request.urlretrieve(
            'https://s3.amazonaws.com/mountebank/v1.4/mountebank-v1.4.3-linux-x64.tar.gz')
        mb_tar = tarfile.open(mb_archive_path, mode='r:gz')
        mb_tar.extractall(self.CACHE_DIR)

        os.remove(mb_archive_path)

    def _check_mb_install(self):
        """
        Checks if Mountebank is normally installed in the system.
        :return: True if Mountebank is installed, False otherwise.
        :rtype: bool
        """
        try:
            subprocess.check_call(
                self.MB_INSTALL_CHECK_CMD,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL)
            self._log.debug('Detected normal Mountebank installation.')
            return True
        except FileNotFoundError:
            self._log.debug('No normal Mountebank installation found.')
            return False

    @staticmethod
    def _get_mb_dir(cache_dir):
        try:
            mb_dir = next(path for path in os.listdir(cache_dir) if path.startswith('mountebank-v'))
            return os.path.join(cache_dir, mb_dir)
        except StopIteration:
            raise MBStandaloneBrokenError(
                "No Mountebank distribution found! "
                "It should have been automatically downloaded earlier")

    @staticmethod
    def _get_node_path(mb_dir):
        try:
            node_path = next(path for path in os.listdir(mb_dir) if path.startswith('node-v'))
            return os.path.join(mb_dir, node_path, 'bin/node')
        except StopIteration:
            raise MBStandaloneBrokenError(
                'No NodeJS distribution found in Mountebank distribution! WTF? Distribution dir: '
                + mb_dir)



"""
Abstractions for management of HTTP service processes.
"""

import atexit
import concurrent.futures
import logging
import socket
import subprocess
import time

import port_for


def wait_for_port(port, host='localhost', timeout=5.0):
    """
    Wait until a port starts accepting TCP connections.
    :param int port: Number of the port.
    :param str host: Host address on which the port should exist.
    :param float timeout: In seconds. How long to wait before raising errors.
    :rtype: None
    :raises TimeoutError: The port isn't accepting connection after time specified in `timeout`.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port)):
                break
        except ConnectionRefusedError:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                raise TimeoutError('Waited too long for the port to start accepting connections.')


class HttpService:

    """
    Manages a HTTP service instance on localhost. Can start and stop the process.
    """

    def __init__(self, process_command, port=None, env=None):
        """
        Initializes the object without starting the service process.
        :param process_command: Command that will start the service.
        It is a string or a list of strings (like parameters to subprocess.Popen).
        Command strings may contain '{port}' - it will be filled with the provided port.
        :param int port: Port on which the service will listen.
        If not provided then the service will run on a random free port.
        :param dict env: Environment variables that will be visible for the service process.
        All values of this dictionary should be strings.
        E.g. {'EXAMPLE_VARIABLE_NAME': 'some_example_value'}
        Values may contain '{port}' string - it will be filled with the provided port.
        """
        if port is None:
            self.port = port_for.select_random()
        else:
            self.port = port
        self.base_url = 'http://localhost:{}'.format(self.port)
        self._process_command = self._format_process_command(process_command, self.port)
        self._service_env = self._format_process_env(env, self.port)
        self._service_proc = None

    def start(self, timeout=5.0):
        """
        Starts service process and waits for it to start accepting connections.
        :param float timeout: How long to wait before raising an error.
        :rtype: None
        :raises TimeoutError: If the service process didn't start in time.
        """
        self._service_proc = subprocess.Popen(self._process_command, env=self._service_env)
        atexit.register(self.stop)

        try:
            wait_for_port(self.port, timeout=timeout)
        except Exception:
            logging.exception("Service '%s' didn't start", self._process_command)
            self.stop()
            raise

    def stop(self):
        atexit.unregister(self.stop)
        self._service_proc.terminate()
        # TODO add timeout and some error
        self._service_proc.wait()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @staticmethod
    def _format_process_command(command, port):
        if type(command) != str:
            return [part.format(port=port) if ('{port}' in part) else part for part in command]
        else:
            return command

    @staticmethod
    def _format_process_env(env, port):
        if not env:
            return env

        formatted_env = dict(env)
        for key, value in env.items():
            if '{port}' in value:
                formatted_env[key] = value.format(port=port)
        return formatted_env


class ServiceGroup:

    """
    Manages a group of service processes.
    Can be used to concurrently start or stop more than one service.
    """

    def __init__(self, *service_processes):
        """
        :param service_processes: A list of HttpService objects.
        Don't start or stop them individually after passing them here.
        :raises TimeoutError:
        """
        self._services = service_processes
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=7)

    def start(self, timeout=5.0):
        """
        Starts all the service processes and waits them to start accepting connections.
        :param float timeout: How long to wait before raising an error.
        :rtype: None
        :raises TimeoutError: If all of the services didn't start in time.
        """
        start_futures = [self._executor.submit(service.start) for service in self._services]
        results = concurrent.futures.wait(start_futures, timeout=timeout)
        if results.not_done:
            raise TimeoutError('Not all processes started in time.')

    def stop(self, timeout=5.0):
        """
        Stops all the service processes. Waits for full stop.
        :param float timeout: How long to wait before raising an error.
        :rtype: None
        :raises TimeoutError: If all of the services didn't stop in time.
        """
        stop_futures = [self._executor.submit(service.stop) for service in self._services]
        results = concurrent.futures.wait(stop_futures, timeout=timeout)
        if results.not_done:
            raise TimeoutError('Not all processes stopped in time.')

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



"""
Abstractions representing aspects of Mountebank.
"""

import collections
import time

import port_for
import requests

from mountepy import HttpService
from mountepy.mb_mgmt import get_mb_command


class Imposter:

    """
    Represents a Mountebank imposter.
    """

    def __init__(self, mountebank_port, port):
        self._url = 'http://localhost:{}/imposters/{}'.format(mountebank_port, port)
        self.port = port

    def requests(self):
        """
        :return: The requests made on the impostor.
        :rtype: list[`ImposterRequest`]
        """
        imposter_json = requests.get(self._url).json()
        imposter_requests = []
        for request in imposter_json.get('requests', []):
            imposter_requests.append(ImposterRequest(
                body=request['body'],
                headers=request['headers'],
                method=request['method'],
                path=request['path'],
                query=request['query'],
                request_from=request['requestFrom'],
            ))
        return imposter_requests

    def wait_for_requests(self, count=1, timeout=5.0):
        start_time = time.perf_counter()
        while True:
            received_requests = self.requests()
            if len(received_requests) >= count:
                return received_requests
            else:
                time.sleep(0.01)
                if time.perf_counter() - start_time >= timeout:
                    raise TimeoutError('Waited too long for requests on stub.')

    def destroy(self):
        """
        Deletes this `Imposter` from Mountebank.
        This object cannot be used afterwards.
        """
        requests.delete(self._url)


class ImposterRequest:

    """
    A request made on a Mountebank imposter. Returned by 'Imposter` object.
    """

    def __init__(self, body, headers, method, path, query, request_from):
        """
        :param body: The body of a response. Can be any valid JSON (this includes raw values)
        :param dict headers:
        :param str method:
        :param str path:
        :param dict query:
        :param str request_from:
        """
        # TODO add timestamp field, use dateutil
        self.body = body
        self.headers = headers
        self.method = method
        self.path = path
        self.query = query
        self.request_from = request_from


HttpStub = collections.namedtuple('HttpStub', ['method', 'path', 'status_code', 'response'])
HttpStub.__doc__ = """A configuration for a simple Mountebank impostor, not including the port."""


# TODO add imposter calling_url field (other url should be management_url)
class Mountebank(HttpService):

    """
    Manages Mountebank instance on localhost. Can start and stop the process.
    """

    def __init__(self, port=None):
        """
        Initializes the object without starting the service process.
        :param int port: Port on which Mountebank will listen for impostor configuration commands.
        If not provided then a random free port will be selected.
        """
        super().__init__(get_mb_command() + ['--mock', '--port', '{port}'], port)
        self._imposters_url = 'http://localhost:{}/imposters'.format(self.port)

    def add_imposter(self, imposter_cfg):
        """
        Adds a HTTP service stub (imposter) to Mountebank instance.
        :param dict imposter_cfg: JSON with Mountebank configuration for an impostor.
        Consult Mountebank documentation.
        :return: The created service stub.
        :rtype: Imposter
        """
        resp = requests.post(self._imposters_url, json=imposter_cfg)
        resp.raise_for_status()
        return Imposter(self.port, imposter_cfg['port'])

    def add_imposter_simple(self, port=None, method='GET', path='/', status_code=200, response=''):
        """
        Adds a HTTP service stub (imposter) to Mountebank instance.
        Takes a simplified configuration.
        :param int port: port the imposter will listen on. If none, a random port will be selected.
        :param str method: HTTP method that the imposter will wait for. GET by default.
        :param str path: HTTP path the imposter will wait for. '/' by default.
        :param int status_code: HTTP status code the imposter will return. 200 by default.
        :param str response: body of the imposters response. Empty string by default.
        :return: The created service stub.
        :rtype: Imposter
        """
        if port == None:
            port = port_for.select_random()

        imposter_config = {
            'port': port,
            'protocol': 'http',
            'stubs': [
                {
                    'responses': [
                        {
                            'is': {
                                'statusCode': status_code,
                                'headers': {
                                    'Content-Type': 'application/json'
                                },
                                'body': response
                            }
                        }
                    ],
                    'predicates': [
                        {
                            'and': [
                                {
                                    'equals': {
                                        'path': path,
                                        'method': method,
                                    }
                                },
                            ]
                        }
                    ]
                }
            ]
        }
        return self.add_imposter(imposter_config)

    def add_imposters_simple(self, port, stubs):
        """
        Adds a Mountebank imposter with multiple HTTP stubs on one port.
        Takes a simplified configuration in comparison to "add_imposter".
        :param int port: port the imposter will listen on
        :param list[HttpStub] stubs: HTTP stubs to be created on the port
        :rtype: None
        """
        # TODO make this method and add_imposter_simple more elegant and less duplicated
        imposter_config = {
            'port': port,
            'protocol': 'http',
            'stubs': []
        }

        for stub in stubs:
            stub_json = {
                'responses': [
                    {
                        'is': {
                            'statusCode': stub.status_code,
                            'headers': {
                                'Content-Type': 'application/json'
                            },
                            'body': stub.response
                        }
                    }
                ],
                'predicates': [
                    {
                        'and': [
                            {
                                'equals': {
                                    'path': stub.path,
                                    'method': stub.method,
                                    }
                            },
                        ]
                    }
                ]
            }
            imposter_config['stubs'].append(stub_json)
        self.add_imposter(imposter_config)

    def reset(self):
        """
        Removes configured imposters (HTTP stubs).
        :rtype: None
        """
        # TODO add validation
        requests.delete(self._imposters_url)
